import copy
import json
from typing import Union
from urllib.parse import urlencode

from py_yt.core.constants import (
    requestPayload,
    searchKey,
    ResultMode,
    videoElementKey,
    channelElementKey,
    playlistElementKey,
    shelfElementKey,
    richItemKey,
)
from py_yt.core.requests import RequestCore
from py_yt.handlers.componenthandler import ComponentHandler
from py_yt.handlers.requesthandler import RequestHandler


class SearchCore(RequestCore, RequestHandler, ComponentHandler):
    response = None
    responseSource = None
    resultComponents = []

    def __init__(
        self,
        query: str,
        limit: int,
        language: str,
        region: str,
        searchPreferences: str,
        timeout: int,
    ):
        super().__init__()
        self.query = query
        self.limit = limit
        self.language = language
        self.region = region
        self.searchPreferences = searchPreferences
        self.timeout = timeout
        self.continuationKey = None

    def sync_create(self):
        self._makeRequest()
        self._parseSource()

    def _getRequestBody(self):
        """Fixes #47"""
        requestBody = copy.deepcopy(requestPayload)
        requestBody["query"] = self.query
        requestBody["client"] = {
            "hl": self.language,
            "gl": self.region,
        }
        if self.searchPreferences:
            requestBody["params"] = self.searchPreferences
        if self.continuationKey:
            requestBody["continuation"] = self.continuationKey
        self.url = (
            "https://www.youtube.com/youtubei/v1/search"
            + "?"
            + urlencode(
                {
                    "key": searchKey,
                }
            )
        )
        self.data = requestBody

    def _makeRequest(self) -> None:
        self._getRequestBody()
        request = self.syncPostRequest()
        try:
            self.response = request.text
        except:
            raise Exception("ERROR: Could not make request.")

    async def _makeAsyncRequest(self) -> None:
        self._getRequestBody()
        request = await self.asyncPostRequest()
        try:
            self.response = request.text
        except:
            raise Exception("ERROR: Could not make request.")

    def result(self, mode: int = ResultMode.dict) -> Union[str, dict]:
        """Returns the search result.

        Args:
            mode (int, optional): Sets the type of result. Defaults to ResultMode.dict.

        Returns:
            Union[str, dict]: Returns JSON or dictionary.
        """
        if mode == ResultMode.json:
            return json.dumps({"result": self.resultComponents}, indent=4)
        elif mode == ResultMode.dict:
            return {"result": self.resultComponents}

    def _next(self) -> bool:
        """Gets the subsequent search result. Call result

        Args:
            mode (int, optional): Sets the type of result. Defaults to ResultMode.dict.

        Returns:
            Union[str, dict]: Returns True if getting more results was successful.
        """
        if self.continuationKey:
            self.response = None
            self.responseSource = None
            self.resultComponents = []
            self._makeRequest()
            self._parseSource()
            self._getComponents(*self.searchMode)
            return True
        else:
            return False

    async def _nextAsync(self) -> dict:
        self.response = None
        self.responseSource = None
        self.resultComponents = []
        await self._makeAsyncRequest()
        self._parseSource()
        self._getComponents(*self.searchMode)
        return {
            "result": self.resultComponents,
        }

    def _getComponents(
        self, findVideos: bool, findChannels: bool, findPlaylists: bool
    ) -> None:
        self.resultComponents = []
        for element in self.responseSource:
            if videoElementKey in element.keys() and findVideos:
                self.resultComponents.append(self._getVideoComponent(element))
            if channelElementKey in element.keys() and findChannels:
                self.resultComponents.append(self._getChannelComponent(element))
            if playlistElementKey in element.keys() and findPlaylists:
                self.resultComponents.append(self._getPlaylistComponent(element))
            if shelfElementKey in element.keys() and findVideos:
                for shelfElement in self._getShelfComponent(element)["elements"]:
                    self.resultComponents.append(
                        self._getVideoComponent(
                            shelfElement,
                            shelfTitle=self._getShelfComponent(element)["title"],
                        )
                    )
            if richItemKey in element.keys() and findVideos:
                richItemElement = self._getValue(element, [richItemKey, "content"])
                """ Initial fallback handling for VideosSearch """
                if videoElementKey in richItemElement.keys():
                    videoComponent = self._getVideoComponent(richItemElement)
                    self.resultComponents.append(videoComponent)
            if len(self.resultComponents) >= self.limit:
                break



import json
from typing import Iterable, Mapping, TypeVar, Union, List

from py_yt.core.componenthandler import getVideoId, getValue
from py_yt.core.constants import searchKey, ResultMode
from py_yt.core.requests import RequestCore

K = TypeVar("K")
T = TypeVar("T")


class CommentsCore(RequestCore):
    result = None
    continuationKey = None
    isNextRequest = False
    response = None

    def __init__(self, video_link: str):
        super().__init__()
        self.commentsComponent = {"result": []}
        self.responseSource = None
        self.videoLink = video_link

    def prepare_continuation_request(self):
        self.data = {
            "context": {
                "client": {"clientName": "WEB", "clientVersion": "2.20210820.01.00"}
            },
            "videoId": getVideoId(self.videoLink),
        }
        self.url = f"https://www.youtube.com/youtubei/v1/next?key={searchKey}"

    def prepare_comments_request(self):
        self.data = {
            "context": {
                "client": {"clientName": "WEB", "clientVersion": "2.20210820.01.00"}
            },
            "continuation": self.continuationKey,
        }

    def parse_source(self):
        self.responseSource = getValue(
            self.response.json(),
            [
                "onResponseReceivedEndpoints",
                0 if self.isNextRequest else 1,
                (
                    "appendContinuationItemsAction"
                    if self.isNextRequest
                    else "reloadContinuationItemsCommand"
                ),
                "continuationItems",
            ],
        )

    def parse_continuation_source(self):
        self.continuationKey = getValue(
            self.response.json(),
            [
                "contents",
                "twoColumnWatchNextResults",
                "results",
                "results",
                "contents",
                -1,
                "itemSectionRenderer",
                "contents",
                0,
                "continuationItemRenderer",
                "continuationEndpoint",
                "continuationCommand",
                "token",
            ],
        )

    async def async_make_comment_request(self):
        self.prepare_comments_request()
        self.response = await self.asyncPostRequest()
        if self.response.status_code == 200:
            self.parse_source()

    async def async_make_continuation_request(self):
        self.prepare_continuation_request()
        self.response = await self.asyncPostRequest()
        if self.response.status_code == 200:
            self.parse_continuation_source()
            if not self.continuationKey:
                raise Exception("Could not retrieve continuation token")
        else:
            raise Exception(
                "Status code is not 200 but " + str(self.response.status_code)
            )

    async def async_create(self):
        await self.async_make_continuation_request()
        await self.async_make_comment_request()
        self.__getComponents()

    async def async_create_next(self):
        self.isNextRequest = True
        await self.async_make_comment_request()
        self.__getComponents()

    def __getComponents(self) -> None:
        comments = []
        for comment in self.responseSource:
            comment = getValue(
                comment, ["commentThreadRenderer", "comment", "commentRenderer"]
            )
            # print(json.dumps(comment, indent=4))
            try:
                j = {
                    "id": self.__getValue(comment, ["commentId"]),
                    "author": {
                        "id": self.__getValue(
                            comment, ["authorEndpoint", "browseEndpoint", "browseId"]
                        ),
                        "name": self.__getValue(comment, ["authorText", "simpleText"]),
                        "thumbnails": self.__getValue(
                            comment, ["authorThumbnail", "thumbnails"]
                        ),
                    },
                    "content": self.__getValue(
                        comment, ["contentText", "runs", 0, "text"]
                    ),
                    "published": self.__getValue(
                        comment, ["publishedTimeText", "runs", 0, "text"]
                    ),
                    "isLiked": self.__getValue(comment, ["isLiked"]),
                    "authorIsChannelOwner": self.__getValue(
                        comment, ["authorIsChannelOwner"]
                    ),
                    "voteStatus": self.__getValue(comment, ["voteStatus"]),
                    "votes": {
                        "simpleText": self.__getValue(
                            comment, ["voteCount", "simpleText"]
                        ),
                        "label": self.__getValue(
                            comment,
                            [
                                "voteCount",
                                "accessibility",
                                "accessibilityData",
                                "label",
                            ],
                        ),
                    },
                    "replyCount": self.__getValue(comment, ["replyCount"]),
                }
                comments.append(j)
            except:
                pass

        self.commentsComponent["result"].extend(comments)
        self.continuationKey = self.__getValue(
            self.responseSource,
            [
                -1,
                "continuationItemRenderer",
                "continuationEndpoint",
                "continuationCommand",
                "token",
            ],
        )

    def __result(self, mode: int) -> Union[dict, str]:
        if mode == ResultMode.dict:
            return self.commentsComponent
        elif mode == ResultMode.json:
            return json.dumps(self.commentsComponent, indent=4)

    def __getValue(
        self, source: dict, path: Iterable[str]
    ) -> Union[str, int, dict, None]:
        value = source
        for key in path:
            if type(key) is str:
                if key in value.keys():
                    value = value[key]
                else:
                    value = None
                    break
            elif type(key) is int:
                if len(value) != 0:
                    value = value[key]
                else:
                    value = None
                    break
        return value

    def __getAllWithKey(self, source: Iterable[Mapping[K, T]], key: K) -> Iterable[T]:
        for item in source:
            if key in item:
                yield item[key]

    def __getValueEx(
        self, source: dict, path: List[str]
    ) -> Iterable[Union[str, int, dict, None]]:
        if len(path) <= 0:
            yield source
            return
        key = path[0]
        upcoming = path[1:]
        if key is None:
            following_key = upcoming[0]
            upcoming = upcoming[1:]
            if following_key is None:
                raise Exception(
                    "Cannot search for a key twice consecutive or at the end with no key given"
                )
            values = self.__getAllWithKey(source, following_key)
            for val in values:
                yield from self.__getValueEx(val, path=upcoming)
        else:
            val = self.__getValue(source, path=[key])
            yield from self.__getValueEx(val, path=upcoming)

    def __getFirstValue(
        self, source: dict, path: Iterable[str]
    ) -> Union[str, int, dict, None]:
        values = self.__getValueEx(source, list(path))
        for val in values:
            if val is not None:
                return val
        return None



from typing import Union


def getValue(source: dict, path: list[str]) -> Union[str, int, dict, None]:
    value = source
    for key in path:
        if type(key) is str:
            if key in value.keys():
                value = value[key]
            else:
                value = None
                break
        elif type(key) is int:
            if len(value) != 0:
                value = value[key]
            else:
                value = None
                break
    return value


def getVideoId(video_link: str) -> str:
    if "youtu.be" in video_link:
        if video_link[-1] == "/":
            return video_link.split("/")[-2]
        return video_link.split("/")[-1]
    elif "youtube.com" in video_link:
        if "&" not in video_link:
            return video_link[video_link.index("v=") + 2 :]
        return video_link[video_link.index("v=") + 2 : video_link.index("&")]
    else:
        return video_link



from .extras import Video, Playlist, Suggestions, Hashtag, Comments, Transcript, Channel
from .search import (
    Search,
    VideosSearch,
    ChannelsSearch,
    PlaylistsSearch,
    CustomSearch,
    ChannelSearch,
)

from .handlers import ComponentHandler, RequestHandler

__all__ = [
    "Video",
    "Playlist",
    "Suggestions",
    "Hashtag",
    "Comments",
    "Transcript",
    "Channel",
    "Search",
    "VideosSearch",
    "ChannelsSearch",
    "PlaylistsSearch",
    "CustomSearch",
    "ChannelSearch",
    "ComponentHandler",
    "RequestHandler",
]



import copy
import json
from typing import Union
from urllib.parse import urlencode
from urllib.request import Request, urlopen

import httpx

from py_yt.core.constants import (
    videoElementKey,
    ResultMode,
    requestPayload,
    searchKey,
    userAgent,
    contentPath,
    hashtagElementKey,
    hashtagBrowseKey,
    hashtagVideosPath,
    hashtagContinuationVideosPath,
    richItemKey,
    continuationKeyPath,
)
from py_yt.handlers.componenthandler import ComponentHandler


class HashtagCore(ComponentHandler):
    response = None
    resultComponents = []

    def __init__(
        self, hashtag: str, limit: int, language: str, region: str, timeout: int
    ):
        self.hashtag = hashtag
        self.limit = limit
        self.language = language
        self.region = region
        self.timeout = timeout
        self.continuationKey = None
        self.params = None

    def result(self, mode: int = ResultMode.dict) -> Union[str, dict]:
        """Returns the hashtag videos.
        Args:
            mode (int, optional): Sets the type of result. Defaults to ResultMode.dict.
        Returns:
            Union[str, dict]: Returns JSON or dictionary.
        """
        if mode == ResultMode.json:
            return json.dumps({"result": self.resultComponents}, indent=4)
        elif mode == ResultMode.dict:
            return {"result": self.resultComponents}

    def next(self) -> bool:
        """Gets the videos from the next page. Call result
        Returns:
            bool: Returns True if getting more results was successful.
        """
        self.response = None
        self.resultComponents = []
        if self.continuationKey:
            self._makeRequest()
            self._getComponents()
        if self.resultComponents:
            return True
        return False

    def _getParams(self) -> None:
        requestBody = copy.deepcopy(requestPayload)
        requestBody["query"] = "#" + self.hashtag
        requestBody["client"] = {
            "hl": self.language,
            "gl": self.region,
        }
        requestBodyBytes = json.dumps(requestBody).encode("utf_8")
        request = Request(
            "https://www.youtube.com/youtubei/v1/search"
            + "?"
            + urlencode(
                {
                    "key": searchKey,
                }
            ),
            data=requestBodyBytes,
            headers={
                "Content-Type": "application/json; charset=utf-8",
                "Content-Length": len(requestBodyBytes),
                "User-Agent": userAgent,
            },
        )
        try:
            response = urlopen(request, timeout=self.timeout).read().decode("utf_8")
        except:
            raise Exception("ERROR: Could not make request.")
        content = self._getValue(json.loads(response), contentPath)
        for item in self._getValue(content, [0, "itemSectionRenderer", "contents"]):
            if hashtagElementKey in item.keys():
                self.params = self._getValue(
                    item[hashtagElementKey],
                    ["onTapCommand", "browseEndpoint", "params"],
                )
                return

    async def _asyncGetParams(self) -> None:
        requestBody = copy.deepcopy(requestPayload)
        requestBody["query"] = "#" + self.hashtag
        requestBody["client"] = {
            "hl": self.language,
            "gl": self.region,
        }
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    "https://www.youtube.com/youtubei/v1/search",
                    params={
                        "key": searchKey,
                    },
                    headers={
                        "User-Agent": userAgent,
                    },
                    json=requestBody,
                    timeout=self.timeout,
                )
                response = response.json()
        except:
            raise Exception("ERROR: Could not make request.")
        content = self._getValue(response, contentPath)
        for item in self._getValue(content, [0, "itemSectionRenderer", "contents"]):
            if hashtagElementKey in item.keys():
                self.params = self._getValue(
                    item[hashtagElementKey],
                    ["onTapCommand", "browseEndpoint", "params"],
                )
                return

    def _makeRequest(self) -> None:
        if self.params == None:
            return
        requestBody = copy.deepcopy(requestPayload)
        requestBody["browseId"] = hashtagBrowseKey
        requestBody["params"] = self.params
        requestBody["client"] = {
            "hl": self.language,
            "gl": self.region,
        }
        if self.continuationKey:
            requestBody["continuation"] = self.continuationKey
        requestBodyBytes = json.dumps(requestBody).encode("utf_8")
        request = Request(
            "https://www.youtube.com/youtubei/v1/browse"
            + "?"
            + urlencode(
                {
                    "key": searchKey,
                }
            ),
            data=requestBodyBytes,
            headers={
                "Content-Type": "application/json; charset=utf-8",
                "Content-Length": len(requestBodyBytes),
                "User-Agent": userAgent,
            },
        )
        try:
            self.response = (
                urlopen(request, timeout=self.timeout).read().decode("utf_8")
            )
        except:
            raise Exception("ERROR: Could not make request.")

    async def _asyncMakeRequest(self) -> None:
        if self.params == None:
            return
        requestBody = copy.deepcopy(requestPayload)
        requestBody["browseId"] = hashtagBrowseKey
        requestBody["params"] = self.params
        requestBody["client"] = {
            "hl": self.language,
            "gl": self.region,
        }
        if self.continuationKey:
            requestBody["continuation"] = self.continuationKey
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    "https://www.youtube.com/youtubei/v1/browse",
                    params={
                        "key": searchKey,
                    },
                    headers={
                        "User-Agent": userAgent,
                    },
                    json=requestBody,
                    timeout=self.timeout,
                )
                self.response = response.content
        except:
            raise Exception("ERROR: Could not make request.")

    def _getComponents(self) -> None:
        if self.response == None:
            return
        self.resultComponents = []
        try:
            if not self.continuationKey:
                responseSource = self._getValue(
                    json.loads(self.response), hashtagVideosPath
                )
            else:
                responseSource = self._getValue(
                    json.loads(self.response), hashtagContinuationVideosPath
                )
            if responseSource:
                for element in responseSource:
                    if richItemKey in element.keys():
                        richItemElement = self._getValue(
                            element, [richItemKey, "content"]
                        )
                        if videoElementKey in richItemElement.keys():
                            videoComponent = self._getVideoComponent(richItemElement)
                            self.resultComponents.append(videoComponent)
                    if len(self.resultComponents) >= self.limit:
                        break
                self.continuationKey = self._getValue(
                    responseSource[-1], continuationKeyPath
                )
        except:
            raise Exception("ERROR: Could not parse YouTube response.")



import json
from typing import Union
from urllib.parse import urlencode

from py_yt.core.constants import ResultMode
from py_yt.core.requests import RequestCore


class SuggestionsCore(RequestCore):
    """Gets search suggestions for the given query.

    Args:
        language (str, optional): Sets the suggestion language. Defaults to 'en'.
        region (str, optional): Sets the suggestion region. Defaults to 'US'.

    Examples:
        Calling `result` method gives the search result.

        >>> suggestions = Suggestions(language = 'en', region = 'US').get('Harry Styles', mode = ResultMode.json)
        >>> print(suggestions)
        {
            'result': [
                'harry styles',
                'harry styles treat people with kindness',
                'harry styles golden music video',
                'harry styles interview',
                'harry styles adore you',
                'harry styles watermelon sugar',
                'harry styles snl',
                'harry styles falling',
                'harry styles tpwk',
                'harry styles sign of the times',
                'harry styles jingle ball 2020',
                'harry styles christmas',
                'harry styles live',
                'harry styles juice'
            ]
        }
    """

    def __init__(self, language: str = "en", region: str = "US", timeout: int = None):
        super().__init__()
        self.language = language
        self.region = region
        self.timeout = timeout

    def _post_request_processing(self, mode):
        searchSuggestions = []

        self.__parseSource()
        for element in self.responseSource:
            if type(element) is list:
                for searchSuggestionElement in element:
                    searchSuggestions.append(searchSuggestionElement[0])
                break
        if mode == ResultMode.dict:
            return {"result": searchSuggestions}
        elif mode == ResultMode.json:
            return json.dumps({"result": searchSuggestions}, indent=4)

    async def _getAsync(
        self, query: str, mode: int = ResultMode.dict
    ) -> Union[dict, str]:
        self.url = (
            "https://clients1.google.com/complete/search"
            + "?"
            + urlencode(
                {
                    "hl": self.language,
                    "gl": self.region,
                    "q": query,
                    "client": "youtube",
                    "gs_ri": "youtube",
                    "ds": "yt",
                }
            )
        )

        await self.__makeAsyncRequest()
        return self._post_request_processing(mode)

    def __parseSource(self) -> None:
        try:
            self.responseSource = json.loads(
                self.response[self.response.index("(") + 1 : self.response.index(")")]
            )
        except:
            raise Exception("ERROR: Could not parse YouTube response.")

    async def __makeAsyncRequest(self) -> None:
        request = await self.asyncGetRequest()
        self.response = request.text
